--- erts/emulator/sys/unix/sys_float.c.orig	Wed Dec 15 00:37:29 2004
+++ erts/emulator/sys/unix/sys_float.c	Wed Jan 26 13:22:04 2005
@@ -172,10 +172,15 @@
 
 #endif
 
-#if defined(__linux__) && defined(__x86_64__)
+#if (defined(__linux__) || defined(__FreeBSD__)) && defined(__x86_64__)
+#include <ucontext.h>
 static void skip_sse2_insn(mcontext_t *mc)
 {
+#if defined(__linux__)
     unsigned char *pc0 = (unsigned char*)(mc->gregs[REG_RIP]);
+#elif defined(__FreeBSD__)
+    unsigned char *pc0 = (unsigned char*)(mc->mc_rip);
+#endif
     unsigned char *pc = pc0;
     unsigned int opcode;
     unsigned int nr_skip_bytes;
@@ -267,13 +272,21 @@
        no need to check the 15-byte instruction length limit here. */
 
     /* Done. */
+#if defined(__linux__)
     mc->gregs[REG_RIP] = (long)pc;
+#elif defined(__FreeBSD__)
+    mc->mc_rip = (__register_t)pc;
+#endif
 }
-#endif /* __linux__ && __x86_64__ */
+#endif /* (__linux__ || __FreeBSD__) && __x86_64__ */
 
-#if (defined(__linux__) && (defined(__i386__) || defined(__x86_64__) || defined(__powerpc__))) || defined(__DARWIN__)
+#if (defined(__linux__) && (defined(__i386__) || defined(__x86_64__) || defined(__powerpc__))) || defined(__DARWIN__) || (defined(__FreeBSD__) && defined(__x86_64__))
 
 #include <ucontext.h>
+#if (defined(__FreeBSD__) && defined(__x86_64__))
+#include <sys/types.h>
+#include <machine/fpu.h>
+#endif
 
 static void fpe_sig_action(int sig, siginfo_t *si, void *puc)
 {
@@ -307,6 +320,20 @@
     mcontext_t mc = uc->uc_mcontext;
     mc->ss.srr0 += 4;
     mc->fs.fpscr = 0x80|0x40|0x10;
+#elif (defined(__FreeBSD__) && defined(__x86_64__))
+    mcontext_t *mc = &uc->uc_mcontext;
+    struct envxmm *ex = (struct envxmm *)mc->mc_fpstate;
+    /* A failed SSE2 instruction will restart. To avoid
+       looping, we must update RIP to skip the instruction
+       (leaving garbage in the destination).
+       The alternative is to mask SSE2 exceptions now and
+       unmask them again later in erts_check_fpe(), but that
+       relies too much on other code being cooperative. */
+    if( ex->en_mxcsr & 0x000F ) {
+	ex->en_mxcsr &= ~(0x003F|0x0780);
+	skip_sse2_insn(mc);
+    }
+    ex->en_sw &= ~0xFF;
 #endif
     erl_fp_exception = 1;
 }
@@ -321,7 +348,7 @@
     unmask_fpe();
 }
 
-#else  /* !((__linux__ && (__i386__ || __x86_64__ || __powerpc__)) || __DARWIN__) */
+#else  /* !((__linux__ && (__i386__ || __x86_64__ || __powerpc__)) || __DARWIN__) || (__FreeBSD__ && __x64_64__) */
 
 static void fpe_sig_handler(int sig)
 {
@@ -335,7 +362,7 @@
     unmask_fpe();
 }
 
-#endif /* (__linux__ && (__i386__ || __x86_64__ || __powerpc__)) || __DARWIN__ */
+#endif /* (__linux__ && (__i386__ || __x86_64__ || __powerpc__)) || __DARWIN__ || (__FreeBSD__ && __x64_64__) */
 
 #endif /* NO_FPE_SIGNALS */
 
